import * as THREE from 'three';
import * as HangarScene from './scene.js';
import * as Utils from './utils.js';

var showLight = true; // You can show the light for debugging

// Global variables, as before. We also have another camera - the reflection camera.
var renderer, scene, camera, reflectionCamera;
var chopper;

// We need to use another fragment shader for the reflective floor
var phongVertexShader, phongFragmentShader, phongReflectiveFragmentShader;

// We also need a texture and a specific material for our floor.
var floorReflectionTexture;
var floorReflectiveMaterial;
var floor;

// The lookAt point is the same for both cameras
var lookAt = new THREE.Vector3(0.0, -10.0, -10.0);

// Let's use a point for the light's position.
var lightPosition = new THREE.Vector3();
var lightTrajectory;
var light;

// Another point for the viewer's position.
var viewerPosition = new THREE.Vector3(0.0, 0.0, 8.0);

window.onLoad = async () => { 
    var canvasContainer = document.getElementById('myCanvasContainer');
    var width = 800;
    var height = 500;

    // Our normal shaders
    phongVertexShader = await(await fetch('phong_vertex.glsl')).text();
    phongFragmentShader = await(await fetch('phong_fragment.glsl')).text();
    
    // The reflective shader for our floor
    phongReflectiveFragmentShader = await(await fetch('phong_reflect_fragment.glsl')).text();

    // Setup
    renderer = new THREE.WebGLRenderer();
    renderer.setSize(width, height);
    canvasContainer.appendChild(renderer.domElement);

    scene = new THREE.Scene();

    // We will use a perspective camera, with FOV 80 degrees, correct aspect ratio, near plane at distance 1, far plane at distance 1000
    camera = new THREE.PerspectiveCamera(80, width / height, 1, 1000);
    camera.position.set(viewerPosition.x, viewerPosition.y, viewerPosition.z);
    camera.up = new THREE.Vector3(0,1,0);
    camera.lookAt(lookAt);

    // We use a camera with the same parameters to capture the reflection.
    reflectionCamera = new THREE.PerspectiveCamera(80, width / height, 1, 1000);

    // --Task--
    // Setup the reflectionCamera properly: correct position, up-vector and lookAt values.


    // We create a texture to store our reflection rendering to.
    floorReflectionTexture = new THREE.WebGLRenderTarget(width, height, {
        minFilter: THREE.LinearFilter,
        magFilter: THREE.NearestFilter,
        format: THREE.RGBAFormat 
    });

    // For debugging purposes you might want to render a light source as a red cube
    // Everyone likes red cubes :)
    if (showLight) {
        var geometry = new THREE.BoxGeometry(0.2, 0.2, 0.2);
        var material = new THREE.MeshBasicMaterial( {color: 'red'} );
        light = new THREE.Mesh(geometry, material);
        scene.add(light);
    }

    // Add the hangar and the chopper
    scene.add(HangarScene.createHangar());
    chopper = HangarScene.createChopper();
    scene.add(chopper);

    // You can get objects from the scene by name, if you have named them.
    floor = scene.getObjectByName('hangar').getObjectByName('floor');

    // Let's create a reflective material for our floor.
    floorReflectiveMaterial = createReflectiveShaderMaterial(new THREE.Color(0x222222));

    // Overwrite the floor's material with our reflective material.
    // Also set the texture value for the material.
    floor.material = floorReflectiveMaterial;

    // --Task-- Add this line after you have specified the shader uniforms
    // floor.material.uniforms.reflectionTexture.value = floorReflectionTexture;


    // Light travels in a cubic spline curve.
    lightTrajectory = new THREE.CatmullRomCurve3([
        new THREE.Vector3( 0, -4, 5 ),
        new THREE.Vector3( -2, 0, 3 ),
        new THREE.Vector3( -3, 4, 6 ),
        new THREE.Vector3( 8, 2, 9 ),
    ]);
    lightTrajectory.closed = true;

    draw();

    // --Task-- do different things on key presses
    window.addEventListener('keydown', function(event) {
        if (event.code == "ArrowUp") { //Up arrow

            // Swap the cameras, so that you can see what is seen from the reflection camera

        }

        if (event.code == "ArrowRight") { //Right arrow
            changeLookAt(true);
        }

        if (event.code == "ArrowLeft") { //Left arrow
            changeLookAt(false);
        }
    }, false);
}

/**
 *  --Task--
 *  This function should change the point the cameras are looking at.
 *  Depending on the dir value, move the point 1 unit or -1 unit along the z-axis.
 */
function changeLookAt(dir) {

    // if (dir) { ... } else { ... }
    // update lookAt variable, update cameras

}

function draw() {
    requestAnimationFrame(draw);

    var m = Utils.millis() / 4000;
    lightPosition = lightTrajectory.getPoint(m - parseInt(m)); // Update the light's position

    if (showLight) { // If we show the light, update the position of the red cube as well
        light.position.set(lightPosition.x, lightPosition.y, lightPosition.z);
    }

    // Just a way rotate the blades and move the chopper up and down. Feel free to copy your own idea here.
    chopper.getObjectByName('blades').rotation.set(0, Utils.toRad((Utils.millis() / 10) % 360), 0);
    chopper.position.set(0, 3.0 * Math.cos(Math.PI * 2 * (Utils.millis() / 25 % 360) / 360) - 5.0, 0);


    // -- 1. Pass - We render with the reflection camera into the storage texture --

    // First we need to find out the light's position in our reflected camera space and send it for all the object that use our shader
    scene.traverse(function(object) {
        if (object.material !== undefined && object.material instanceof THREE.ShaderMaterial) { // We have our own shaders for this object
            // --Task--
            // Assign the lightPosition in the reflection camera's space
            // object.material.uniforms.lightPosition.value = ...
            // Cameras in Three.js have the .matrixWorldInverse matrix stored in them.
        }
    });

    // Next we disable the floor. The reflection camera doesn't see it anyway.
    floor.visible = false;

    // We render the scene with our reflection camera and store the result in a texture.
    renderer.setRenderTarget(floorReflectionTexture);
    renderer.render(scene, reflectionCamera);
    renderer.setRenderTarget(null);

    // -- 2. Pass - We render with the main camera.

    // Enable the foor.
    floor.visible = true;

    // We need to recalculate the light's position again. This time it should be in the main camera's space.
    scene.traverse(function(object) {
        if (object.material !== undefined && object.material instanceof THREE.ShaderMaterial) {
            // --Task--
            // Do the same as before, but take the view matrix from the main camera.
        }
    });

    // Normal rendering into the screen buffer.
    renderer.render(scene, camera);
}

export function createShaderMaterial(color) {
			
    return new THREE.ShaderMaterial({
        uniforms: {
            // --Task--
            // Send two uniforms to the shader. This is the same as before.
            // -lightPosition, v3, vec3 - the position of the light source
            // -color, c, vec3 - the color of the current material
        },
        vertexShader: phongVertexShader,
        fragmentShader: phongFragmentShader
    });
}


function createReflectiveShaderMaterial(color) {
    return new THREE.ShaderMaterial({
        uniforms: {
            // --Task--
            // Send three uniforms to the shader. Light position and color are the same,
            // but we also need to send the texture that we rendered to in the first pass.
            // -lightPosition, v3, vec3 - the position of the light source
            // -color, c, vec3 - the color of the current material
            // -reflectionTexture, t, sampler2D
        },
        vertexShader: phongVertexShader,
        fragmentShader: phongReflectiveFragmentShader // We use the different fragment shader for this material
    });
}